// GCLIB (Free GameCube SDK Libraries)
//
// GC_BBA.h for GC_BBA.c by Costis
//
// Version 1.0
//


#ifndef GC_BBA_H
#define GC_BBA_H

extern unsigned char ghwaddr[6];
extern unsigned char shwaddr[6];
extern unsigned char BBArevID;

extern unsigned char pdata[2048];

#pragma pack(1)

typedef struct {
	unsigned char shaddress[6];
	unsigned char dhaddress[6];
	unsigned short type;
} ETHpacket;

typedef struct {
	unsigned short htype;
	unsigned short ptype;
	unsigned char halen;
	unsigned char palen;
	unsigned short opcode;
	unsigned char shaddress[6];
	unsigned char spaddress[4];
	unsigned char dhaddress[6];
	unsigned char dpaddress[4];
} ARPpacket;

typedef struct {
	unsigned version:4;
	unsigned IHL:4;
	unsigned char TOS;
	unsigned short len;
	unsigned short ID;
	unsigned flags:3;
	unsigned foffset:13;
	unsigned char TTL;
	unsigned char protocol;
	unsigned short checksum;
	unsigned long srcIP;
	unsigned long dstIP;
	//unsigned long options;
} IPpacket;

typedef struct {
	unsigned short srcPort;
	unsigned short dstPort;
	unsigned short len;
	unsigned short checksum;
} UDPpacket;

#pragma pack(0)

#define GC_BBA_COMMAND_WRITE		0xC0000000
#define GC_BBA_COMMAND_READ			0x80000000

#define GC_BBA_COMMAND_WRITES		0x4000
#define GC_BBA_COMMAND_READS		0x0000

unsigned char GC_BBA_ReceiveCommand (unsigned long reg, unsigned char *abuf, unsigned long size);
unsigned char GC_BBA_SendCommand (unsigned long reg, unsigned char *abuf, unsigned long size);
unsigned char GC_BBA_ReceiveCommandShort (unsigned long reg, unsigned char *abuf, unsigned long size);
unsigned char GC_BBA_SendCommandShort (unsigned long reg, unsigned char *abuf, unsigned long size);

void GC_BBA_Reset (void);
unsigned char GC_BBA_Init (void);
unsigned char GC_BBA_Send (unsigned char *abuf, unsigned short size);
unsigned short GC_BBA_Receive (unsigned char *abuf);
unsigned short GC_BBA_GenerateChecksum (unsigned short *abuf, unsigned long hsize);

#define GC_BBA_ARP_REQUEST		1
#define GC_BBA_ARP_REPLY		2

void GC_BBA_GenerateARP (ARPpacket *tARP, unsigned short type, unsigned long srcIP, unsigned long dstIP);
int GC_BBA_ParseARP (ARPpacket *tARP, char *hwaddr);
int GC_BBA_SendARP (unsigned char *hwaddr, unsigned long srcIP, unsigned long dstIP);
void GC_BBA_GenerateIP (IPpacket *tIP, unsigned long srcIP, unsigned long dstIP, unsigned short length);
void GC_BBA_GenerateUDP (UDPpacket *tUDP, unsigned long srcPort, unsigned long dstPort, unsigned short length);

#define IPtol(a,b,c,d)	(((a) << 24) | ((b) << 16) | ((c) << 8) | (d))
#define PCIP		IPtol (192,168,1,100)
#define GCIP		IPtol (192,168,1,15) // This should just be any unused IP.
#define PCport		9002
#define GCport		1234

int GC_BBA_SendUDP (unsigned char *abuf, unsigned short size);
int GC_BBA_SendUDPE (unsigned short tport, unsigned char *abuf, unsigned short size);
void GC_BBA_RString (char *astr);

#endif
