#ifndef __GC_CONTEXT_H__
#define __GC_CONTEXT_H__

//#include <GC_Types.h>

// Floating point context modes
#define     GC_CONTEXT_MODE_FPU         0x01u
#define     GC_CONTEXT_MODE_PSFP        0x02u

// Context status
#define     GC_CONTEXT_STATE_FPSAVED    0x01u   // set if FPU is saved
#define     GC_CONTEXT_STATE_EXC        0x02u   // set if saved by exception

typedef struct
{
    // General-purpose registers
    unsigned long GPR[32];

    unsigned long CR;
    unsigned long LR;
    unsigned long CTR;
    unsigned long XER;

    // Floating-point registers
    double FPR[32];

    unsigned long FPSCR_pad;
    unsigned long FPSCR;

    // Exception handling registers
    unsigned long SRR0;
    unsigned long SRR1;

    // Context mode
    unsigned short mode;           // since UIMM is 16 bits in PPC
    unsigned short state;          // OR-ed OS_CONTEXT_STATE_*

    // Place Gekko regs at the end so we have minimal changes to
    // existing code
    unsigned long GQR[8];
    unsigned long PSF_pad;
    double PSF[32];

} GC_Context;

#define __GC_CONTEXT_FRAME  768

#define GC_CONTEXT_R0       0
#define GC_CONTEXT_R1       4
#define GC_CONTEXT_R2       8
#define GC_CONTEXT_R3       12
#define GC_CONTEXT_R4       16
#define GC_CONTEXT_R5       20
#define GC_CONTEXT_R6       24
#define GC_CONTEXT_R7       28
#define GC_CONTEXT_R8       32
#define GC_CONTEXT_R9       36
#define GC_CONTEXT_R10      40
#define GC_CONTEXT_R11      44
#define GC_CONTEXT_R12      48
#define GC_CONTEXT_R13      52
#define GC_CONTEXT_R14      56
#define GC_CONTEXT_R15      60
#define GC_CONTEXT_R16      64
#define GC_CONTEXT_R17      68
#define GC_CONTEXT_R18      72
#define GC_CONTEXT_R19      76
#define GC_CONTEXT_R20      80
#define GC_CONTEXT_R21      84
#define GC_CONTEXT_R22      88
#define GC_CONTEXT_R23      92
#define GC_CONTEXT_R24      96
#define GC_CONTEXT_R25      100
#define GC_CONTEXT_R26      104
#define GC_CONTEXT_R27      108
#define GC_CONTEXT_R28      112
#define GC_CONTEXT_R29      116
#define GC_CONTEXT_R30      120
#define GC_CONTEXT_R31      124

#define GC_CONTEXT_CR       128
#define GC_CONTEXT_LR       132
#define GC_CONTEXT_CTR      136
#define GC_CONTEXT_XER      140

#define GC_CONTEXT_FPR0     144
#define GC_CONTEXT_FPR1     152
#define GC_CONTEXT_FPR2     160
#define GC_CONTEXT_FPR3     168
#define GC_CONTEXT_FPR4     176
#define GC_CONTEXT_FPR5     184
#define GC_CONTEXT_FPR6     192
#define GC_CONTEXT_FPR7     200
#define GC_CONTEXT_FPR8     208
#define GC_CONTEXT_FPR9     216
#define GC_CONTEXT_FPR10    224
#define GC_CONTEXT_FPR11    232
#define GC_CONTEXT_FPR12    240
#define GC_CONTEXT_FPR13    248
#define GC_CONTEXT_FPR14    256
#define GC_CONTEXT_FPR15    264
#define GC_CONTEXT_FPR16    272
#define GC_CONTEXT_FPR17    280
#define GC_CONTEXT_FPR18    288
#define GC_CONTEXT_FPR19    296
#define GC_CONTEXT_FPR20    304
#define GC_CONTEXT_FPR21    312
#define GC_CONTEXT_FPR22    320
#define GC_CONTEXT_FPR23    328
#define GC_CONTEXT_FPR24    336
#define GC_CONTEXT_FPR25    344
#define GC_CONTEXT_FPR26    352
#define GC_CONTEXT_FPR27    360
#define GC_CONTEXT_FPR28    368
#define GC_CONTEXT_FPR29    376
#define GC_CONTEXT_FPR30    384
#define GC_CONTEXT_FPR31    392

#define GC_CONTEXT_FPSCR    400     // 8 bytes including padding

#define GC_CONTEXT_SRR0     408
#define GC_CONTEXT_SRR1     412

#define GC_CONTEXT_MODE     416     // only 2 bytes
#define GC_CONTEXT_STATE    418     // only 2 bytes

#define GC_CONTEXT_GQR0     420
#define GC_CONTEXT_GQR1     424
#define GC_CONTEXT_GQR2     428
#define GC_CONTEXT_GQR3     432
#define GC_CONTEXT_GQR4     436
#define GC_CONTEXT_GQR5     440
#define GC_CONTEXT_GQR6     444
#define GC_CONTEXT_GQR7     448
#define __GCCONTEXT_PADDING 452     // double word alignment for the 64 bit psf

#define GC_CONTEXT_PSF0     456
#define GC_CONTEXT_PSF1     464
#define GC_CONTEXT_PSF2     472
#define GC_CONTEXT_PSF3     480
#define GC_CONTEXT_PSF4     488
#define GC_CONTEXT_PSF5     496
#define GC_CONTEXT_PSF6     504
#define GC_CONTEXT_PSF7     512
#define GC_CONTEXT_PSF8     520
#define GC_CONTEXT_PSF9     528
#define GC_CONTEXT_PSF10    536
#define GC_CONTEXT_PSF11    544
#define GC_CONTEXT_PSF12    552
#define GC_CONTEXT_PSF13    560
#define GC_CONTEXT_PSF14    568
#define GC_CONTEXT_PSF15    576
#define GC_CONTEXT_PSF16    584
#define GC_CONTEXT_PSF17    592
#define GC_CONTEXT_PSF18    600
#define GC_CONTEXT_PSF19    608
#define GC_CONTEXT_PSF20    616
#define GC_CONTEXT_PSF21    624
#define GC_CONTEXT_PSF22    632
#define GC_CONTEXT_PSF23    640
#define GC_CONTEXT_PSF24    648
#define GC_CONTEXT_PSF25    656
#define GC_CONTEXT_PSF26    664
#define GC_CONTEXT_PSF27    672
#define GC_CONTEXT_PSF28    680
#define GC_CONTEXT_PSF29    688
#define GC_CONTEXT_PSF30    696
#define GC_CONTEXT_PSF31    704

#define GC_CONTEXT_CURRENT			*(unsigned long*)(0x800000D4)
#define GC_CONTEXT_EXCEPTIONVECTOR	*(unsigned long*)(0x800000C0)

extern GC_Context *GC_DefaultContext;

void GC_Context_Init (void);
void GC_Context_Load (GC_Context *tcontext);
void GC_Context_Save (GC_Context *tcontext);

#endif
