#ifndef __GC_INTERRUPT_H__
#define __GC_INTERRUPT_H__

#include <GC_Types.h>
#include <GC_Exception.h>
#include <GC_Context.h>

typedef s16  GC_Interrupt;
typedef void (*GC_Interrupt_Handler)(GC_Interrupt interrupt, GC_Context* context);
typedef u32  GC_Interrupt_Mask;

#define INTERRUPT_CAUSE			*(u32*)0xCC003000

#define INTERRUPT_MASK			(*(u32*)0x800000C8)

#define INTERRUPT_CAUSE_CACHE	*(u32*)0x800000C4
#define INTERRUPT_MASK_CACHE	*(u32*)0x800000C8


#define GC_INTERRUPT_MEM_0                0
#define GC_INTERRUPT_MEM_1                1
#define GC_INTERRUPT_MEM_2                2
#define GC_INTERRUPT_MEM_3                3
#define GC_INTERRUPT_MEM_ADDRESS          4
#define GC_INTERRUPT_DSP_AI               5
#define GC_INTERRUPT_DSP_ARAM             6
#define GC_INTERRUPT_DSP_DSP              7
#define GC_INTERRUPT_AI_AI                8
#define GC_INTERRUPT_EXI_0_EXI            9
#define GC_INTERRUPT_EXI_0_TC             10
#define GC_INTERRUPT_EXI_0_EXT            11
#define GC_INTERRUPT_EXI_1_EXI            12
#define GC_INTERRUPT_EXI_1_TC             13
#define GC_INTERRUPT_EXI_1_EXT            14
#define GC_INTERRUPT_EXI_2_EXI            15
#define GC_INTERRUPT_EXI_2_TC             16
#define GC_INTERRUPT_PI_CP                17
#define GC_INTERRUPT_PI_PE_TOKEN          18
#define GC_INTERRUPT_PI_PE_FINISH         19
#define GC_INTERRUPT_PI_SI                20
#define GC_INTERRUPT_PI_DI                21
#define GC_INTERRUPT_PI_RSW               22
#define GC_INTERRUPT_PI_ERROR             23
#define GC_INTERRUPT_PI_VI                24
#define GC_INTERRUPT_PI_DEBUG             25
#define GC_INTERRUPT_PI_HSP               26
#define GC_INTERRUPT_MAX                  32

#define GC_INTERRUPTMASK(interrupt)         (0x80000000u >> (interrupt))        // for cntlzw

extern volatile int resetGC;

void GC_Interrupts_Init(void);
u32  GC_Interrupts_Disable(void);
u32  GC_Interrupts_Enable(void);
u32  GC_Interrupts_Restore(u32 state);
void GC_Interrupts_SetMask (unsigned long interruptMask);
void GC_Interrupts_MaskInternal (unsigned long interruptMask);
void GC_Interrupts_UnmaskInternal (unsigned long interruptMask);
extern __GC_Exception_Handler GC_Interrupts_ExternalInterruptHandler (u32 var1, u32 var2, u32 var3, u32 var4);

#endif
