// GCLIB (Free GameCube SDK Libraries)
//
// GC_BBA.c by Costis
//
// Version 1.0
//
// Revisions:
//
// V1.0 - Added all of the current functions, etc.
//		  Note: They are all synchronous and do not
//				use any interrupts.
//
//
//	Currently, these libraries only have UDP support,
//	as the TCP protocol is far more complicated
//	protocol and I have not yet had the chance to
//	implement it.

#include <GC_BBA.h>
#include <GC_Types.h>
#include <GC_Debug.h>
#include <GC_EXI.h>

unsigned char ghwaddr[6];
unsigned char shwaddr[6];
unsigned char BBArevID;

unsigned char pdata[2048];

unsigned char GC_BBA_ReceiveCommand (unsigned long reg, unsigned char *abuf, unsigned long size)
{
	unsigned long aval;

	GC_EXI_Select (0, 2, 5);

	// Send the read command with the specific register to the BBA. 
	aval = GC_BBA_COMMAND_READ | (reg << 8);
	GC_EXI_Imm (0, (unsigned char*)&aval, 4, EXI_CONTROL_TYPE_WRITE);

	// Receive the data 
	GC_EXI_ImmEx (0, abuf, size, EXI_CONTROL_TYPE_READ);

	GC_EXI_Deselect (0);
}

unsigned char GC_BBA_SendCommand (unsigned long reg, unsigned char *abuf, unsigned long size)
{
	unsigned long aval;

	GC_EXI_Select (0, 2, 5);

	// Send the read command with the specific register to the BBA. 
	aval = GC_BBA_COMMAND_WRITE | (reg << 8);
	GC_EXI_Imm (0, (unsigned char*)&aval, 4, EXI_CONTROL_TYPE_WRITE);

	// Receive the data 
	GC_EXI_ImmEx (0, abuf, size, EXI_CONTROL_TYPE_WRITE);

	GC_EXI_Deselect (0);
}

unsigned char GC_BBA_ReceiveCommandShort (unsigned long reg, unsigned char *abuf, unsigned long size)
{
	unsigned short aval;

	GC_EXI_Select (0, 2, 5);

	// Send the read command with the specific register to the BBA. 
	aval = GC_BBA_COMMAND_READS | (reg << 8);
	GC_EXI_Imm (0, (unsigned char*)&aval, 2, EXI_CONTROL_TYPE_WRITE);

	// Receive the data 
	GC_EXI_ImmEx (0, abuf, size, EXI_CONTROL_TYPE_READ);

	GC_EXI_Deselect (0);
}

unsigned char GC_BBA_SendCommandShort (unsigned long reg, unsigned char *abuf, unsigned long size)
{
	unsigned short aval;

	GC_EXI_Select (0, 2, 5);

	// Send the read command with the specific register to the BBA. 
	aval = GC_BBA_COMMAND_WRITES | (reg << 8);
	GC_EXI_Imm (0, (unsigned char*)&aval, 2, EXI_CONTROL_TYPE_WRITE);

	// Receive the data 
	GC_EXI_ImmEx (0, abuf, size, EXI_CONTROL_TYPE_WRITE);

	GC_EXI_Deselect (0);
}

unsigned char GC_BBA_Send (u8 *abuf, u16 size)
{
	u8 getbuf, aval;
	u32 i;

	if (size < 60) 
		size = 60;

	GC_BBA_SendCommand (0x48, abuf, size);

	GC_BBA_ReceiveCommand (0x00,&getbuf,1);
	getbuf |= 4;
	aval = 0;
	GC_BBA_SendCommand (0x00,&aval,1);
	GC_BBA_SendCommand (0x00,&getbuf,1);

	//for (i = 0; i < 0x100000; i++);
	while ((volatile)(getbuf) & 0x06) 
		GC_BBA_ReceiveCommand (0x00,&getbuf,1);

	return 1;
}


unsigned short GC_BBA_Receive (unsigned char *abuf)
{
	unsigned char aval;
	unsigned short len;
	unsigned short p_read, p_write;
	unsigned char descr[4];

	GC_BBA_ReceiveCommand (0x16, (unsigned char*)&p_write, 2);
	GC_BBA_ReceiveCommand (0x18, (unsigned char*)&p_read, 2);
	
	if (p_read == p_write)
		return 0;

	aval = 0x02;
	GC_BBA_SendCommand (0x3A, &aval, 1);
	GC_BBA_ReceiveCommand (0x3A, &aval, 1);

	if (aval & 2)
		return 0;
	
	GC_BBA_ReceiveCommand (p_read, descr, 4);
	len = (descr[1] >> 4) | (descr[2] << 4);

	if ((p_read + 4 + len) <= 0x1000)
		GC_BBA_ReceiveCommand (p_read + 4, abuf, len);
	else {
		aval = 0x1000 - (p_read + 4);
		GC_BBA_ReceiveCommand (p_read + 4, abuf, aval);
		GC_BBA_ReceiveCommand (0x100, abuf + aval, len - aval);
	}

	descr[1] = 0;
	GC_BBA_SendCommand (0x18, descr, 2);

	aval = 0x02;
	GC_BBA_SendCommand (0x09, &aval, 1);

	aval = 0xF8;
	GC_BBA_SendCommandShort (0x02, &aval, 1);

	return len;
}

unsigned char GC_BBA_Init (void)
{
	int i;
	unsigned char a;
	unsigned short b;

	a = 0;
	GC_BBA_SendCommand (0x60, (unsigned char*)&a, 1);	// unknown
	for (i = 0; i < 4; i++)
		GC_Video_WaitForVBlank ();
	GC_BBA_ReceiveCommand (0xF, &a, 1);
	for (i = 0; i < 4; i++)
		GC_Video_WaitForVBlank ();
	a = 1;
	GC_BBA_SendCommand (0, (unsigned char*)&a, 1);
	for (i = 0; i < 4; i++)
		GC_Video_WaitForVBlank ();
	a = 0;
	GC_BBA_SendCommand (0, (unsigned char*)&a, 1);
	
	GC_BBA_ReceiveCommand (1, (unsigned char*)&a, 1);
	
	GC_BBA_ReceiveCommand (0x5B, (unsigned char*)&a, 1);
	a &= ~(1<<7);
	GC_BBA_SendCommand (0x5B, (unsigned char*)&a, 1);	


	a = 1;
	GC_BBA_SendCommand (0x5E, (unsigned char*)&a, 1);	
	
	GC_BBA_ReceiveCommand (0x5C, (unsigned char*)&a, 1);
	a |= 4;
	GC_BBA_SendCommand (0x5C, (unsigned char*)&a, 1);	

	a = 0x11;
	GC_BBA_SendCommand (1, (unsigned char*)&a, 1);
	a = 0x80;
	GC_BBA_SendCommand (0x50, (unsigned char*)&a, 1);	
	a = 0xFF;
	GC_BBA_SendCommand (8, (unsigned char*)&a, 1);
	GC_BBA_SendCommand (9, (unsigned char*)&a, 1);
	a = 0;
	GC_BBA_SendCommand (2, (unsigned char*)&a, 1);
	a = 8;
	GC_BBA_SendCommand (0, (unsigned char*)&a, 1);

	for (i = 0; i < 4; i++)
		GC_Video_WaitForVBlank ();

	// Receive function initialization.

	b = 0x0100; // Starting position of the receive buffer.
	GC_BBA_SendCommand (0x16, (unsigned char*)&b, 2);
	GC_BBA_SendCommand (0x18, (unsigned char*)&b, 2);

	// Read the MAC address.
	GC_BBA_ReceiveCommand (0x20, shwaddr, 6);

	return 1;
}

unsigned short GC_BBA_GenerateChecksum (unsigned short *abuf, unsigned long hsize)
{
	int i;
	unsigned long cval;

	cval = 0;
	for (i = 0; i < hsize; i++)
	{
		cval += abuf[i];

		cval = (cval & 0xFFFF) + (cval >> 16);
		cval = (cval & 0xFFFF) + (cval >> 16);		
	}

	return (unsigned short)(cval ^ 0xFFFF);
}

unsigned long pID = 0;

void GC_BBA_GenerateARP (ARPpacket *tARP, unsigned short type, unsigned long srcIP, unsigned long dstIP)
{
	tARP->htype = 1; // Ethernet.
	tARP->ptype = 0x800; // IP.
	tARP->halen = 6;
	tARP->palen = 4;
	tARP->opcode = type;

	// Source MAC Address.
	tARP->shaddress[0] = 0x00; 
	tARP->shaddress[1] = 0x00; 
	tARP->shaddress[2] = 0x00; 
	tARP->shaddress[3] = 0x00; 
	tARP->shaddress[4] = 0x00; 
	tARP->shaddress[5] = 0x00; 

	tARP->spaddress[0] = srcIP >> 24;
	tARP->spaddress[1] = (srcIP >> 16) & 0xFF;
	tARP->spaddress[2] = (srcIP >> 8) & 0xFF;
	tARP->spaddress[3] = srcIP & 0xFF;

	// Destination MAC Address.
	tARP->dhaddress[0] = 0x00; 
	tARP->dhaddress[1] = 0x00; 
	tARP->dhaddress[2] = 0x00; 
	tARP->dhaddress[3] = 0x00; 
	tARP->dhaddress[4] = 0x00; 
	tARP->dhaddress[5] = 0x00; 

	tARP->dpaddress[0] = dstIP >> 24;
	tARP->dpaddress[1] = (dstIP >> 16) & 0xFF;
	tARP->dpaddress[2] = (dstIP >> 8) & 0xFF;
	tARP->dpaddress[3] = dstIP & 0xFF;
}

int GC_BBA_ParseARP (ARPpacket *tARP, char *hwaddr)
{
	// There should be some more ARP packet checking code here.
	if (tARP->opcode == 2)
	{
		hwaddr[0] = tARP->shaddress[0];
		hwaddr[1] = tARP->shaddress[1];
		hwaddr[2] = tARP->shaddress[2];
		hwaddr[3] = tARP->shaddress[3];
		hwaddr[4] = tARP->shaddress[4];
		hwaddr[5] = tARP->shaddress[5];
		return 1;
	} else
		return 0; // Unimplemented ARP packet type.
}

int GC_BBA_SendARPR (unsigned long srcIP, unsigned long dstIP)
{
	ARPpacket tARP, *iARP;

	GC_Memory_memset (pdata, 0, 42);

	// Destination MAC Address.
	pdata[0] = ghwaddr[0];
	pdata[1] = ghwaddr[1];
	pdata[2] = ghwaddr[2];
	pdata[3] = ghwaddr[3];
	pdata[4] = ghwaddr[4];
	pdata[5] = ghwaddr[5];
	
	// Source MAC Address.
	pdata[6] = 0x00;
	pdata[7] = 0x00;
	pdata[8] = 0x00;
	pdata[9] = 0x00;
	pdata[10] = 0x00;
	pdata[11] = 0x00;

	pdata[12] = 0x08;
	pdata[13] = 0x06;

	GC_BBA_GenerateARP (&tARP, GC_BBA_ARP_REPLY, srcIP, dstIP);

	tARP.dhaddress[0] = ghwaddr[0]; 
	tARP.dhaddress[1] = ghwaddr[1]; 
	tARP.dhaddress[2] = ghwaddr[2]; 
	tARP.dhaddress[3] = ghwaddr[3]; 
	tARP.dhaddress[4] = ghwaddr[4];
	tARP.dhaddress[5] = ghwaddr[5]; 
	
	GC_Memory_memcpy (pdata + 14, (unsigned char*)&tARP, 28);
	GC_Memory_DCFlushRange (pdata, 32768);

	GC_BBA_Send (pdata, 42);

	return 1;
}

int GC_BBA_SendARP (unsigned char *hwaddr, unsigned long srcIP, unsigned long dstIP)
{
	ARPpacket tARP, *iARP;
	unsigned long i, len;

	GC_Memory_memset (pdata, 0, 42);

	// Destination MAC Address.
	pdata[0] = 0xFF;
	pdata[1] = 0xFF;
	pdata[2] = 0xFF;
	pdata[3] = 0xFF;
	pdata[4] = 0xFF;
	pdata[5] = 0xFF;
	
	// Source MAC Address.
	pdata[6] = 0x00;
	pdata[7] = 0x00;
	pdata[8] = 0x00;
	pdata[9] = 0x00;
	pdata[10] = 0x00;
	pdata[11] = 0x00;

	pdata[12] = 0x08;
	pdata[13] = 0x06;

	GC_BBA_GenerateARP (&tARP, GC_BBA_ARP_REQUEST, srcIP, dstIP);
	
	GC_Memory_memcpy (pdata + 14, (unsigned char*)&tARP, 28);
	GC_Memory_DCFlushRange (pdata, 32768);

	GC_BBA_Send (pdata, 42);

	// This should be changed so that there is a receive data time-out.
	while (1)
	{
		i = 0x10000000;
		while (!(len = GC_BBA_Receive (pdata)) || (!i))
			i--;
		if (!i)
			return -1;
		//printf ("Packet Size: %d, Packet Type: %X", len, ((ETHpacket*)pdata)->type);

		if (((ETHpacket*)pdata)->type == 0x806)
			break;
	}

	//printf ("Out of ARP receive loop!\n");

	iARP = (ARPpacket*)(pdata + 14);

	return GC_BBA_ParseARP (iARP, hwaddr);
}

void GC_BBA_GenerateIP (IPpacket *tIP, unsigned long srcIP, unsigned long dstIP, unsigned short length)
{
	int i;
	unsigned short *cbuf;
	unsigned long cval;

	tIP->version = 4; // IP
	tIP->IHL = 5; // Length of header in words.
	tIP->TOS = 0;
	tIP->len = length; // Total length of headers.
	tIP->ID = pID;
	pID = (pID + 1) & 0xFFFF;
	tIP->flags = 0;
	tIP->foffset = 0;
	tIP->TTL = 0x80;
	tIP->protocol = 0x11; // UDP.
	tIP->checksum = 0;
	tIP->srcIP = srcIP; // Source IP Address.
	tIP->dstIP = dstIP; // Destination IP Address.

	tIP->checksum = GC_BBA_GenerateChecksum ((unsigned short*)tIP, 10);
}

void GC_BBA_GenerateUDP (UDPpacket *tUDP, unsigned long srcPort, unsigned long dstPort, unsigned short length)
{
	tUDP->srcPort = srcPort;
	tUDP->dstPort = dstPort;
	tUDP->len = length;
	tUDP->checksum = 0;
}

int GC_BBA_SendUDP (unsigned char *abuf, unsigned short size)
{
	//unsigned char *pdata;
	IPpacket tIP;
	UDPpacket tUDP;

	//pdata = 0x80600000;
	GC_Memory_memset (pdata, 0, 42 + size);

	// Destination MAC Address.
	pdata[0] = ghwaddr[0];
	pdata[1] = ghwaddr[1];
	pdata[2] = ghwaddr[2];
	pdata[3] = ghwaddr[3];
	pdata[4] = ghwaddr[4];
	pdata[5] = ghwaddr[5];
	
	// Source MAC Address.
	pdata[6] = 0x00;
	pdata[7] = 0x00;
	pdata[8] = 0x00;
	pdata[9] = 0x00;
	pdata[10] = 0x00;
	pdata[11] = 0x00;

	pdata[12] = 0x08;
	pdata[13] = 0x00;

	GC_BBA_GenerateIP (&tIP, GCIP, PCIP, 28 + size);
	GC_BBA_GenerateUDP (&tUDP, GCport, PCport, 8 + size); 
	
	GC_Memory_memcpy (pdata + 14, &tIP, 20);
	GC_Memory_memcpy (pdata + 34, &tUDP, 8);
	GC_Memory_memcpy (pdata + 42, abuf, size);

	GC_Memory_DCFlushRange (pdata, 32768);

	//GC_BBA_Send (pdata, 42 + size);
	return GC_BBA_Send (pdata, 42 + size);
}

int GC_BBA_SendUDPE (unsigned short tport, unsigned char *abuf, unsigned short size)
{
	//unsigned char *pdata;
	IPpacket tIP;
	UDPpacket tUDP;

	//pdata = 0x80600000;
	GC_Memory_memset (pdata, 0, 42 + size);

	// Destination MAC Address.
	pdata[0] = ghwaddr[0];
	pdata[1] = ghwaddr[1];
	pdata[2] = ghwaddr[2];
	pdata[3] = ghwaddr[3];
	pdata[4] = ghwaddr[4];
	pdata[5] = ghwaddr[5];

	// Source MAC Address.
	pdata[6] = 0x00;
	pdata[7] = 0x00;
	pdata[8] = 0x00;
	pdata[9] = 0x00;
	pdata[10] = 0x00;
	pdata[11] = 0x00;

	pdata[12] = 0x08;
	pdata[13] = 0x00;

	GC_BBA_GenerateIP (&tIP, GCIP, PCIP, 28 + size);
	GC_BBA_GenerateUDP (&tUDP, GCport, tport, 8 + size); 
	
	GC_Memory_memcpy (pdata + 14, &tIP, 20);
	GC_Memory_memcpy (pdata + 34, &tUDP, 8);
	GC_Memory_memcpy (pdata + 42, abuf, size);

	GC_Memory_DCFlushRange (pdata, 32768);

	return GC_BBA_Send (pdata, 42 + size);
}

void GC_BBA_RString (char *astr)
{
	unsigned long i;
	
	i = 0;
	while (1)
		if (!astr[i++])
			break;

	GC_BBA_SendUDPE (9003, astr, i);
}
