// GCLIB (Free GameCube SDK Libraries)
//
// GC_Debug.c by Costis
//
// Version 1.0
//
// Revisions:
//
// V1.0 - Added all of the current functions, etc.
//		  Various debug functions have been implemented.

#include <GC_Debug.h>
#include <GC_Video.h>

extern unsigned char font[ ];

unsigned long *Dfbuf; // Debug Frame-Buffer
int Dvmode; // Debug Video Mode

void GC_Debug_SetInstructionBreakpoint (unsigned long address);
void GC_Debug_SetDataBreakpoint (unsigned long address);

	asm (".globl GC_Debug_SetInstructionBreakpoint");
	asm ("GC_Debug_SetInstructionBreakpoint:");
	asm ("mfmsr r4"); // Set the processor state to supervisor mode.
	asm ("rlwinm r5, r4, 0, 18, 16");
	asm ("mtmsr r5");
	asm ("mtspr 1010, r3"); // Set IABR to the instruction break-point address.
	asm ("mtmsr r4"); // Restore the processor state from supervisor mode to whatever mode it was in before.
	asm ("blr");

	asm (".globl GC_Debug_SetDataBreakpoint");
	asm ("GC_Debug_SetDataBreakpoint:");
	asm ("mfmsr r4"); // Set the processor state to supervisor mode.
	asm ("rlwinm r5, r4, 0, 18, 16");
	asm ("mtmsr r5");
	asm ("mtspr 1013, r3"); // Set IABR to the instruction break-point address.
	asm ("mtmsr r4"); // Restore the processor state from supervisor mode to whatever mode it was in before.
	asm ("blr");

void GC_Debug_Init (unsigned long *FrameBuffer, int vmode)
{
	Dfbuf = FrameBuffer;
	Dvmode = vmode;
}

// These font functions can definitely be optimized.
void GC_Debug_RCharT (unsigned char tchar, unsigned long color, int x, int y)
{
	int a, b;

	x >>= 1;
	for (b = 0; b < 16; b += 2)
	{
		for (a = 0; a < 8; a++)
		{
			if (font[(tchar << 6) + a + (b << 2)])
			{
				Dfbuf[x + a + ((y + b) * 320)] = color;
				Dfbuf[x + a + ((y + b) * 320) + 320] = color;
			}
		}
	}
}

void GC_Debug_RHNumT (unsigned long anum, unsigned long color, int x, int y)
{
	int i;
	unsigned char ac;

	for (i = 0; i < 8; i++)
	{
		ac = (unsigned char)((anum >> (28 - (i << 2))) & 0xF);
		if (ac > 9) ac += ('A' - '0' - 10);
		ac += '0';
		
		GC_Debug_RCharT (ac, color, x + (i << 4), y);
	}
}

void GC_Debug_RNumT (long anum, unsigned long color, int x, int y)
{
	int i, j;
	unsigned char tbuf[10];

	if (anum < 0)
	{
		// Negative number!
		GC_Debug_RCharT ('-', color, x, y);
		x += 16;
		anum = -anum;
	}

	i = 0;
	if (anum == 0)
	{
		tbuf[i++] = '0';
	} else {
		while (anum != 0)
		{
			tbuf[i++] = (int)(anum % 10) + '0';
			anum /= 10;
		}
	}

	for (j = i - 1; j >= 0; j--)
		GC_Debug_RCharT (tbuf[j], color, x + (((i - j) - 1) << 4), y);
}

void GC_Debug_RStringT (char *astr, unsigned long color, int x, int y)
{
	int i;
	unsigned char ac;

	i = 0;
	while (*astr != 0)
	{
		GC_Debug_RCharT (*astr++, color, x + i, y);
		i += 16;
	}
}

void GC_Debug_ASSERT (int condition, char *astr, char *expression, char *file, unsigned long tline)
{
	if (!condition)
	{
		GC_Video_Init (Dvmode);
		GC_Video_SetFrameBuffer (Dfbuf, VI_FRAMEBUFFER_BOTH);
		GC_Video_ClearFrameBuffer (Dfbuf, 0x525A52F0);
		
		GC_Debug_RStringT ("GameCube ASSERT Failure!", 0xFF80FF80, 128, 32);
		
		GC_Debug_RStringT ("Message:", 0xFF80FF80, 56, 128);
		GC_Debug_RStringT (astr, 0xFF80FF80, 56, 160);

		GC_Debug_RStringT ("Condition:", 0xFF80FF80, 56, 196);
		GC_Debug_RStringT (expression, 0xFF80FF80, 56, 228);

		GC_Debug_RStringT ("File:", 0xFF80FF80, 56, 264);
		GC_Debug_RStringT (file, 0xFF80FF80, 152, 264);

		GC_Debug_RStringT ("Line: 0x", 0xFF80FF80, 56, 300);
		GC_Debug_RHNumT (tline, 0xFF80FF80, 184, 300);

		GC_Debug_RStringT ("Press START to resume.", 0xFF80FF80, 56, 416);
		GC_Debug_RStringT ("Note: Video recovery is unlikely.", 0xFF80FF80, 56, 432);

		while (1);
	}
}
