// GCLIB (Free GameCube SDK Libraries)
//
// GC_Exception.c by Costis
//
// Version 1.0
//
// Revisions:
//
// V1.0 - Added all of the current functions, etc.
//		  Exception handling should be pretty much
//		  complete! lint helped out with the original
//		  exception code.

#include <GC_Exception.h>
#include <GC_Interrupt.h>
#include <GC_Context.h>
#include <GC_Video.h>
#include <GC_Pad.h>
#include <GC_Debug.h>

#define SPRG0   272

static u32 __GC_Exception_Location[15] = {	0x00000100, 0x00000200, 0x00000300, 0x00000400, \
                                 			0x00000500, 0x00000600, 0x00000700, 0x00000800, \
                                    		0x00000900, 0x00000C00, 0x00000D00, 0x00000F00, \
                                    		0x00001300, 0x00001400, 0x00001700 };

const char *__GC_Exception_Names[15] = {"System Reset", "Machine Check", "DSI", "ISI",
										"Interrupt", "Alignment", "Program", "Floating Point",
										"Decrementer", "System Call", "Trace", "Performance",
										"IABR", "Reserved", "Thermal"};

static u32* __GC_Exception_Table = (u32*) 0x80003000;


unsigned long decCount;

__GC_Exception_Handler GC_Exception_Interrupts (u32 exception, GC_Context *context)
{
	GC_Context_Load (context);
	return;
}

// Unhandled Exception Debugger Function
__GC_Exception_Handler __GC_Exception_Unhandled (u32 exception, GC_Context *context, u32 dsisr, u32 dar)
{
	int i, j;

	// DSI Exception

	GC_Debug_Init ((void*)0xC0500000, 0);
	
	GC_Video_Init (0);
	GC_Video_SetFrameBuffer ((void*)0xC0500000, VI_FRAMEBUFFER_BOTH);
	GC_Video_ClearFrameBuffer ((void*)0xC0500000, 0x525A52F0);

	GC_Debug_RStringT ("GameCube Exception!", 0xFF80FF80, 168, 32);
	GC_Debug_RStringT ("Exception Type:", 0xFF80FF80, 56, 80);
	GC_Debug_RStringT ((char*)__GC_Exception_Names[exception], 0xFF80FF80, 312, 80);

	GC_Debug_RStringT ("Context Information:", 0xFF80FF80, 56, 128);
	GC_Debug_RStringT ("R0:          R1:          R2: ", 0xFF80FF80, 16, 160);
	GC_Debug_RStringT ("R3:          R4:          R5: ", 0xFF80FF80, 16, 160 + 16);
	GC_Debug_RStringT ("R6:          R7:          R8: ", 0xFF80FF80, 16, 160 + 16*2);
	GC_Debug_RStringT ("R9:          R10:         R11:", 0xFF80FF80, 16, 160 + 16*3);
	GC_Debug_RStringT ("R12:         R13:         R14:", 0xFF80FF80, 16, 160 + 16*4);
	GC_Debug_RStringT ("R15:         R16:         R17:", 0xFF80FF80, 16, 160 + 16*5);
	GC_Debug_RStringT ("R18:         R19:         R20:", 0xFF80FF80, 16, 160 + 16*6);
	GC_Debug_RStringT ("R21:         R22:         R23:", 0xFF80FF80, 16, 160 + 16*7);
	GC_Debug_RStringT ("R24:         R25:         R26:", 0xFF80FF80, 16, 160 + 16*8);
	GC_Debug_RStringT ("R27:         R28:         R29:", 0xFF80FF80, 16, 160 + 16*9);
	GC_Debug_RStringT ("R30:         R31:         LR:", 0xFF80FF80, 16, 160 + 16*10);
	GC_Debug_RStringT ("CTR:                     XER:", 0xFF80FF80, 16, 160 + 16*11);
	GC_Debug_RStringT ("SRR0:                    SRR1:", 0xFF80FF80, 16, 160 + 16*12);
	GC_Debug_RStringT ("DISR:                    DAR:", 0xFF80FF80, 16, 160 + 16*13);

	GC_Debug_RHNumT (context->GPR[0], 0xFF80FF80, 80, 160); // R0
	GC_Debug_RHNumT (context->GPR[1], 0xFF80FF80, 80 + 13*16, 160); // R1
	GC_Debug_RHNumT (context->GPR[2], 0xFF80FF80, 80 + 26*16, 160); // R2

	GC_Debug_RHNumT (context->GPR[3], 0xFF80FF80, 80, 160 + 16); // R3
	GC_Debug_RHNumT (context->GPR[4], 0xFF80FF80, 80 + 13*16, 160 + 16); // R4
	GC_Debug_RHNumT (context->GPR[5], 0xFF80FF80, 80 + 26*16, 160 + 16); // R5

	// R6 - R29
	for (i = 2; i < 10; i++)
	{
		GC_Debug_RHNumT (context->GPR[i * 3], 0xFF80FF80, 80, 160 + 16 * i);
		GC_Debug_RHNumT (context->GPR[i * 3 + 1], 0xFF80FF80, 80 + 13*16, 160 + 16 * i);
		GC_Debug_RHNumT (context->GPR[i * 3 + 2], 0xFF80FF80, 80 + 26*16, 160 + 16 * i);
	}

	GC_Debug_RHNumT (context->GPR[30], 0xFF80FF80, 80, 160 + 16*10); // R30
	GC_Debug_RHNumT (context->GPR[31], 0xFF80FF80, 80 + 13*16, 160 + 16*10); // R31

	GC_Debug_RHNumT (context->LR, 0xFF80FF80, 80 + 26*16, 160 + 16*10); // LR
	GC_Debug_RHNumT (context->CTR, 0xFF80FF80, 96, 160 + 16*11); // LR
	GC_Debug_RHNumT (context->XER, 0xFF80FF80, 496, 160 + 16*11); // LR

	GC_Debug_RHNumT (context->SRR0, 0xFF80FF80, 96, 160 + 16*12); // SRR0
	GC_Debug_RHNumT (context->SRR1, 0xFF80FF80, 496, 160 + 16*12); // SRR1
	GC_Debug_RHNumT (*(unsigned long*)context->SRR0, 0xFF80FF80, 96, 160 + 16*13); // SRR0
	GC_Debug_RHNumT (dar, 0xFF80FF80, 496, 160 + 16*13); // SRR1

	GC_Debug_RStringT ("Press START to reset the system.", 0xFF80FF80, 56, 436);

	while(1)
	{
		//j = *(volatile unsigned long*)0xCC006400;
		i = *(volatile unsigned long*)0xCC006404;
		j = *(volatile unsigned long*)0xCC006408;
		if ((i >> 16) & 0x1000)
			*(unsigned long*)(0xCC003024) = 0;
		
		GC_Video_WaitForVBlank ();
		
	}
}

__GC_Exception_Handler __GC_Exception_Test (u32 var1, u32 var2, u32 var3, u32 var4)
{
	int i, j;

	// Default unhandled exception handler.

	GC_Debug_Init ((void*)0xC0500000, 0);
	
	GC_Video_Init (0);
	GC_Video_SetFrameBuffer ((void*)0xC0500000, VI_FRAMEBUFFER_BOTH);
	GC_Video_ClearFrameBuffer ((void*)0xC0500000, 0x525A52F0);

	// Display the various parameters passed to this exception handler.
	GC_Debug_RHNumT (var1, 0xFF80FF80, 32, 32);
	GC_Debug_RHNumT (var2, 0xFF80FF80, 32, 64);
	GC_Debug_RHNumT (var3, 0xFF80FF80, 32, 96);
	GC_Debug_RHNumT (var4, 0xFF80FF80, 32, 128);

	while(1)
	{
		//j = *(volatile unsigned long*)0xCC006400;
		i = *(volatile unsigned long*)0xCC006404;
		j = *(volatile unsigned long*)0xCC006408;
		if ((i >> 16) & 0x1000)
			break; //*(unsigned long*)(0xCC003024) = 0;
		//if ((i >> 16) & 0x0100)
		//	*(unsigned long*)(0xCC003024) = 0;
		
		GC_Video_WaitForVBlank ();
		
	}

	while(1)
	{
		i = *(volatile unsigned long*)0xCC006404;
		j = *(volatile unsigned long*)0xCC006408;
		if (!((i >> 16) & 0x1000))
			break;

		GC_Video_WaitForVBlank ();
		
	}

}

__GC_Exception_Handler TraceException (u32 exception, GC_Context *context)
{
	printf ("Address: %X, Instruction: %X\n", context->SRR0, *(unsigned long*)(context->SRR0));
	GC_Context_Load (context);
}

extern __GC_Exception_Handler __GC_Exception_Handler_Trace(u32 var1, u32 var2, u32 var3, u32 var4);
extern __GC_Exception_Handler __GC_Exception_Handler_Decrementer(u32 var1, u32 var2, u32 var3, u32 var4);
extern __GC_Exception_Handler __GC_Exception_Handler_Default(u32 var1, u32 var2, u32 var3, u32 var4);
extern __GC_Exception_Handler rfiTest(u32 var1, u32 var2, u32 var3, u32 var4);
extern __GC_Exception_Handler GC_Exception_DABRDispatch (u32 var1, u32 var2, u32 var3, u32 var4);

	asm(".globl    __GC_Exception_Vector_Code");

	asm("__GC_Exception_Handler_Default:");
	asm("stw      r0,0(r4)");
    asm("stw      r1,4(r4)");
    asm("stw      r2,8(r4)");
    asm("stmw     r6,24(r4)");
    asm("mfspr    r0,913");
    asm("stw      r0,424(r4)");
    asm("mfspr    r0,914");
    asm("stw      r0,428(r4)");
    asm("mfspr    r0,915");
    asm("stw      r0,432(r4)");
    asm("mfspr    r0,916");
    asm("stw      r0,436(r4)");
    asm("mfspr    r0,917");
    asm("stw      r0,440(r4)");
    asm("mfspr    r0,918");
    asm("stw      r0,444(r4)");
    asm("mfspr    r0,919");
    asm("stw      r0,448(r4)");
	asm("mfdsisr  r5");
    asm("mfdar    r6");
    asm("b        __GC_Exception_Unhandled");


	asm("__GC_Exception_Handler_Trace:");
	asm("stw      r0,0(r4)");
    asm("stw      r1,4(r4)");
    asm("stw      r2,8(r4)");
    asm("stmw     r6,24(r4)");
    asm("mfspr    r0,913");
    asm("stw      r0,424(r4)");
    asm("mfspr    r0,914");
    asm("stw      r0,428(r4)");
    asm("mfspr    r0,915");
    asm("stw      r0,432(r4)");
    asm("mfspr    r0,916");
    asm("stw      r0,436(r4)");
    asm("mfspr    r0,917");
    asm("stw      r0,440(r4)");
    asm("mfspr    r0,918");
    asm("stw      r0,444(r4)");
    asm("mfspr    r0,919");
    asm("stw      r0,448(r4)");
	asm("mfdsisr  r5");
    asm("mfdar    r6");
    asm("b        TraceException");

	asm("__GC_Exception_Handler_Decrementer:");
	asm("stw      r0,0(r4)");
    asm("stw      r1,4(r4)");
    asm("stw      r2,8(r4)");
    asm("stmw     r6,24(r4)");
    asm("mfspr    r0,913");
    asm("stw      r0,424(r4)");
    asm("mfspr    r0,914");
    asm("stw      r0,428(r4)");
    asm("mfspr    r0,915");
    asm("stw      r0,432(r4)");
    asm("mfspr    r0,916");
    asm("stw      r0,436(r4)");
    asm("mfspr    r0,917");
    asm("stw      r0,440(r4)");
    asm("mfspr    r0,918");
    asm("stw      r0,444(r4)");
    asm("mfspr    r0,919");
    asm("stw      r0,448(r4)");
	
	// Increment the Decrementer Exception counter variable.
	asm("lis	  r3, decCount@h");
	asm("ori	  r3, r3, decCount@l");
	asm("lwz	  r5, 0(r3)");
	asm("addi	  r5, r5, 1");
	asm("stw	  r5, 0(r3)");

	// Reset the Decrementer register.
	//asm("lis	  r3, 0xFFFF");
	//asm("ori	  r3, r3, 0xFFFF");
	asm("lis	  r3, 0x0269");
	asm("ori	  r3, r3, 0xFB20");
	asm("mtdec	  r3");

	// Return from the exception.
	asm("mr		  r3, r4");
    asm("b        GC_Context_Load");

	asm(".globl GC_Exception_DABRDispatch");
	asm("GC_Exception_DABRDispatch:");
	asm("stw      r0, 0(r4)");
	asm("stw      r1, 4(r4)");
	asm("stw      r2, 8(r4)");
	asm("stmw     r6, 24(r4)");
	asm("mfspr    r0, 913");
	asm("stw      r0, 424(r4)");
	asm("mfspr    r0, 914");
	asm("stw      r0, 428(r4)");
	asm("mfspr    r0, 915");
	asm("stw      r0, 432(r4)");
	asm("mfspr    r0, 916");
	asm("stw      r0, 436(r4)");
	asm("mfspr    r0, 917");
	asm("stw      r0, 440(r4)");
	asm("mfspr    r0, 918");
	asm("stw      r0, 444(r4)");
	asm("mfspr    r0, 919");
	asm("stw      r0, 448(r4)");
	asm("stwu     r1, -8(r1)");
	asm("mfmsr	  r3");
	asm("ori	  r3, r3, 0x2000");
	asm("mtmsr	  r3");
	asm("mfspr    r5, 18");
	asm("mfspr	  r6, 19");
	asm("b        __GC_Exception_DABR");

extern void	__GC_Exception_Vector_Code(void);
extern u32	__GC_PATCH_VECTOR_CODE;
extern void	__GC_Exception_Vector_End(void);

extern void	__GC_SystemCall_Vector(void);
extern void	__GC_SystemCall_Vector_End(void);

extern void	__GC_Exception_Vector_Code2(void);
extern void	__GC_Exception_Vector_End2(void);

// INLINE ASM
// void __GC_Exception_Vector_Code(void);
//****************************************

	asm(".globl    __GC_Exception_Vector_Code");
	asm(".globl    __GC_PATCH_VECTOR_CODE");
	asm(".globl    __GC_Exception_Vector_End");

    asm("__GC_Exception_Vector_Code:");
    asm("mtspr    272,r4");
    asm("lwz      r4,0x000000C0(0)");
    asm("stw      r3,12(r4)");
    asm("mfspr    r3,272");
    asm("stw      r3,16(r4)");
    asm("stw      r5,20(r4)");
    asm("lhz      r3,418(r4)");
    asm("ori      r3,r3,0x0002");
    asm("sth      r3,418(r4)");
    asm("mfcr     r3");
    asm("stw      r3,128(r4)");
    asm("mflr     r3");
    asm("stw      r3,132(r4)");
    asm("mfctr    r3");
    asm("stw      r3,136(r4)");
    asm("mfxer    r3");
    asm("stw      r3,140(r4)");
    asm("mfsrr0   r3");
    asm("stw      r3,408(r4)");
    asm("mfsrr1   r3");
    asm("stw      r3,412(r4)");
	//asm("b		  __GC_Exception_Test");

    asm("mr       r5,r3");
    asm("nop");
    asm("mfmsr    r3");
    asm("ori      r3,r3,0x0030");
    asm("mtsrr1   r3");
    //asm("stw      r3,412(r4)");
	asm("__GC_PATCH_VECTOR_CODE:");			// we need to patch that part
    asm("li       r3,0");                   // to include excetion number
	asm("lwz      r4, 0x000000D4(0)"); //(r5)");
    asm("rlwinm.  r5,r5,0,30,30");
    asm("bne      __GC_Exception_Vector_Code_Jump");
    asm("lis      r5,__GC_Exception_Handler_Default@ha");
    asm("addi     r5,r5,__GC_Exception_Handler_Default@l");
    asm("mtsrr0   r5");
    asm("rfi");
    asm("__GC_Exception_Vector_Code_Jump:");
    asm("clrlslwi r5,r3,24,2");
    asm("lwz      r5,12288(r5)");
    asm("mtsrr0   r5");
    asm("rfi");
	asm("__GC_Exception_Vector_End:");

	asm(".globl __GC_SystemCall_Vector");
	asm(".globl __GC_SystemCall_Vector_End");
	asm("__GC_SystemCall_Vector:");
	asm("mfspr    r9, 1008");
	asm("ori      r10,r9,0x0008");
	asm("mtspr    1008 ,r10");
	asm("isync");
	asm("sync");
	asm("mtspr    1008, r9");
	asm("rfi");
	asm("__GC_SystemCall_Vector_End:");

void GC_SystemCall_Init (void)
{
	GC_Memory_memcpy (0x80000C00, __GC_SystemCall_Vector, (u32)__GC_SystemCall_Vector_End - (u32)__GC_SystemCall_Vector);
	GC_Memory_DCFlushRangeNoSync (0x80000C00, (u32)__GC_SystemCall_Vector_End - (u32)__GC_SystemCall_Vector);
	GC_Memory_ICInvalidateRange (0x80000C00, (u32)__GC_SystemCall_Vector_End - (u32)__GC_SystemCall_Vector);
}

void GC_Exception_Init (void)
{
	__GC_Exception exceptionNumber = 0;
	u32 *temp;

	// we verify that the exception vector code 
	// is not bigger than 0x100 (MANDATORY cause its OUR code)

	// We copy the 15 handled exception into the start of memory  
    while(exceptionNumber < 0x0f)
    {
	    temp = (u32*) GC_Memory_PhysicalToCached(__GC_Exception_Location[exceptionNumber]);
	    GC_Memory_memcpy(temp, __GC_Exception_Vector_Code, ((u32) __GC_Exception_Vector_End - (u32) __GC_Exception_Vector_Code));
	    GC_Memory_DCFlushRangeNoSync(temp, __GC_Exception_Vector_End - __GC_Exception_Vector_Code);
	    GC_Memory_ICInvalidateRange(temp, __GC_Exception_Vector_End - __GC_Exception_Vector_Code);

		// We need to patch the vector code
		(__GC_PATCH_VECTOR_CODE) = (u32) ((__GC_PATCH_VECTOR_CODE) + 1); //exceptionNumber);

        exceptionNumber++;
    }

	*(unsigned long*)(0x80000100) = 0x48000000;
	GC_Memory_DCFlushRangeNoSync(0x80000100, 64);
	GC_Memory_ICInvalidateRange(0x80000100, 64);

	exceptionNumber = 0;

	// We set the exception handlers 
	// (default one since we init)
    while(exceptionNumber < 0x0f)
    {
        __GC_Exception_Handler_Set(exceptionNumber, (__GC_Exception_Handler) __GC_Exception_Handler_Default);
		exceptionNumber++;
    }

    __GC_Exception_Handler_Set (4, (__GC_Exception_Handler) GC_Interrupts_ExternalInterruptHandler);
    __GC_Exception_Handler_Set (8, (__GC_Exception_Handler) __GC_Exception_Handler_Decrementer);
    //__GC_Exception_Handler_Set (10, (__GC_Exception_Handler) __GC_Exception_Handler_Trace);
}

__GC_Exception_Handler __GC_Exception_Handler_Set(__GC_Exception exceptionNumber, __GC_Exception_Handler handler)
{
	if(exceptionNumber < 15)
	{
		(__GC_Exception_Handler) __GC_Exception_Table[exceptionNumber] = handler;
		return((__GC_Exception_Handler) __GC_Exception_Table[exceptionNumber]);	
	}
	else
	{
        #ifdef GC_DEBUG 
        printf("GC_exception.c : __GC_Exception_Handler_Set : exceptionNumber is not valid (greater than 0x0e).");
        #endif
        return(NULL);
    }
}

__GC_Exception_Handler __GC_Exception_Handler_Get(__GC_Exception exceptionNumber)
{
	if(exceptionNumber < 15)
		return((__GC_Exception_Handler) __GC_Exception_Table[exceptionNumber]);
	else
	{
		#ifdef GC_DEBUG
		printf("GC_exception.c : __GC_Exception_Handler_Get : exceptionNumber is not valid (greater than 0x0e).");
		#endif
		return(NULL);
	}
}
