// GCLIB (Free GameCube SDK Libraries)
//
// GC_Exception.c by Costis
//
// Version 1.0
//
// Revisions:
//
// V1.0 - Added all of the current functions, etc.
//		  Currently there are some simple Video
//		  Interface functions implemented. A
//		  simplistic interrupt handler has also
//		  been implemented, which counts frames.

#include <GC_Video.h>
#include <GC_Types.h>

const u32 GC_VI_MODE_IRegs[VI_MODE_COUNT][32] = {{0x0F060001, 0x476901AD, 0x02EA5140, 0x00030018, // NTSC
								  				     0x00020019, 0x410C410C, 0x40ED40ED, 0x80100000,
													 0x00000000, 0x80100000, 0x00000000, 0x00000000,
													 0x110701AE, 0x10010001, 0x00010001, 0x00010001,
													 0x00000000, 0x00000000, 0x28500100, 0x1AE771F0,
													 0x0DB4A574, 0x00C1188E, 0xC4C0CBE2, 0xFCECDECF,
													 0x13130F08, 0x00080C0F, 0x00FF0000, 0x00000000,
													 0x02800000, 0x000000FF, 0x00FF00FF, 0x00FF00FF},

													{0x11F50101, 0x4B6A01B0, 0x02F85640, 0x00010023, // PAL 50Hz
													 0x00000024, 0x4D2B4D6D, 0x4D8A4D4C, 0x00000000,
													 0x00000000, 0x00000000, 0x00000000, 0x013C0144,
													 0x113901B1, 0x10010001, 0x00010001, 0x00010001,
													 0x00000000, 0x00000000, 0x28500100, 0x1AE771F0,
													 0x0DB4A574, 0x00C1188E, 0xC4C0CBE2, 0xFCECDECF,
													 0x13130F08, 0x00080C0F, 0x00FF0000, 0x00000000,
													 0x02800000, 0x000000FF, 0x00FF00FF, 0x00FF00FF},

													{0x0F060001, 0x476901AD, 0x02EA5140, 0x00030018, // PAL 60Hz
													 0x00020019, 0x410C410C, 0x40ED40ED, 0x00000000,
													 0x00000000, 0x00000000, 0x00000000, 0x00050176,
													 0x110701AE, 0x10010001, 0x00010001, 0x00010001,
													 0x00000000, 0x00000000, 0x28500100, 0x1AE771F0,
								 					 0x0DB4A574, 0x00C1188E, 0xC4C0CBE2, 0xFCECDECF,
													 0x13130F08, 0x00080C0F, 0x00FF0000, 0x00000000,
													 0x02800000, 0x000000FF, 0x00FF00FF, 0x00FF00FF}};

// Initialize the Video Interface Unit
void GC_Video_Init (int TVmode)
{
	int i;

	// Set the VI initialization state.
	for (i = 0; i < 32; i++)
		*(VI_BASE + i) = GC_VI_MODE_IRegs[TVmode][i];

	frameCount = 0;
}

// Set the External Frame-buffer
// 
// fbuf should be an address in low RAM which is 32 byte aligned.
// bnum should be VI_FRAMEBUFFER0, VI_FRAMEBUFFER1, or VI_FRAMEBUFFER_BOTH
void GC_Video_SetFrameBuffer (void *fbuf, int bnum)
{
	if (bnum & VI_FRAMEBUFFER0)
		VI_FBUF1 = (u32)fbuf;
	if (bnum & VI_FRAMEBUFFER1)
	{
		//if (bnum == VI_FRAMEBUFFER_BOTH)
		//	VI_FBUF2 = (u32)fbuf + 640*2;
		//else
			VI_FBUF2 = (u32)fbuf;
	}
}

// Written in PPC assembly and unrolled a bit for optimization.
	asm (".globl GC_Video_ClearFrameBuffer");
	asm ("GC_Video_ClearFrameBuffer:");
	asm ("li	r5, 0x960");
	asm ("mtctr r5");
	asm ("subi	r3, r3, 4");
	asm ("GC_Video_ClearFrameBuffer_CLoop:");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");

	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");
	asm ("stwu	r4, 4(r3)");

	asm ("bdnz	GC_Video_ClearFrameBuffer_CLoop");
	asm ("blr");

unsigned long frameCount;
unsigned long secondCount;

extern unsigned long *cfbuf;

void GC_Video_InterruptHandler (void)
{
	int i;

	if (GC_Video_ClearIF () & 3)
	{
		frameCount++;
		if (!(frameCount % 60))
			secondCount++;
	}
}
	// Some optimized code for clearing the interrupt flags of a VI interrupt.
	// Note: This is an internal library function, please do not call it directly.
	asm (".globl GC_Video_ClearIF");
	asm ("GC_Video_ClearIF:");
	asm ("li	  r3, 0");
	asm ("lis     r9, 0xCC00");
	asm ("ori     r9, r9, 0x2030");
	asm ("lhz     r0, 0(r9)");
	asm ("andi.   r11, r0, 0x8000");
	asm ("beq     aloc_8004AA54");
	asm ("clrlwi  r0, r0, 17");
	asm ("sth     r0, 0(r9)");
	asm ("ori	  r3, r3, 1");

	asm ("aloc_8004AA54:");                   
	asm ("lis     r9, 0xCC00");
	asm ("ori     r9, r9, 0x2034");
	asm ("lhz     r0, 0(r9)");
	asm ("andi.   r11, r0, 0x8000");
	asm ("beq     aloc_8004AA70");
	asm ("clrlwi  r0, r0, 17");
	asm ("sth     r0, 0(r9)");
	asm ("ori	  r3, r3, 2");

	asm ("aloc_8004AA70:");                           
	asm ("lis     r9, 0xCC00");
	asm ("ori     r9, r9, 0x2038");
	asm ("lhz     r0, 0(r9)");
	asm ("andi.   r11, r0, 0x8000");
	asm ("beq     aloc_8004AA8C");
	asm ("clrlwi  r0, r0, 17");
	asm ("sth     r0, 0(r9)");
	asm ("ori	  r3, r3, 4");

	asm ("aloc_8004AA8C:");                        
	asm ("lis     r9, 0xCC00");
	asm ("ori     r9, r9, 0x203C");
	asm ("lhz     r0, 0(r9)");
	asm ("andi.   r11, r0, 0x8000");
	asm ("beqlr");
	asm ("clrlwi  r0, r0, 17");
	asm ("sth     r0, 0(r9)");
	asm ("ori	  r3, r3, 8");
	asm ("blr");

// Synchronous function to wait for the next VBlank.
// When interrupts are enabled, use of the frame
// counter rather than this is recommended.
void GC_Video_WaitForVBlank (void)
{
	while (VI_HLINE != 200); //0x20A);
	while (VI_HLINE <= 200); //0x20A);
}

// Set the TV screen to black!
void GC_Video_SetBlack (unsigned char black)
{
	if (black)
	{
		*(unsigned short*)(0xCC002000) = 0x0006;
		*(unsigned long*)(0xCC00200C) = 0x001501E6;
		*(unsigned long*)(0xCC002010) = 0x001401E7;
	} else { //these settings depend on video mode!!
		*(unsigned short*)(0xCC002000) = 0x0F06;
		*(unsigned long*)(0xCC00200C) = 0x00030018;
		*(unsigned long*)(0xCC002010) = 0x00020019;
	}
}

#define CLAMP(x,l,h) ((x > h) ? h : ((x < l) ? l : x))
#define TENT(a, b, c) (0.25 * (a) + 0.5 * (b) + 0.25 * (c))

#define Yr		0.30078125
#define Yg		0.5859375
#define Yb		0.11328125

#define Cbr		-0.171875
#define Cbg		-0.33984375
#define Cbb		0.51171875

#define Crr		0.51171875
#define Crg		-0.4296875
#define Crb		-0.08203125

// Software RGB to YCBCr conversion code for XFB programs.
unsigned long GC_Video_RGBToYCbCr (unsigned char r, unsigned char g, unsigned char b)
{
	float Y, Cb, Cr;

    Y  =  0.257 * (float)r + 0.504 * (float)g + 0.098 * (float)b +  16.0 + 0.5;
    Cb = -0.148 * (float)r - 0.291 * (float)g + 0.439 * (float)b + 128.0 + 0.5;
    Cr =  0.439 * (float)r - 0.368 * (float)g - 0.071 * (float)b + 128.0 + 0.5;

    Y  = CLAMP(Y , 16, 235);
    Cb = CLAMP(Cb, 16, 240);
    Cr = CLAMP(Cr, 16, 240);

	return (unsigned long)(((unsigned char)Y << 24) | ((unsigned char)Cb << 16) | ((unsigned char)Y << 8) | (unsigned char)Cr);
}
